# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.log import LoggingIntercept
import pyomo.common.unittest as unittest
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.environ import Any, Binary, ConcreteModel, log, Param, Var
from pyomo.repn.parameterized import ParameterizedLinearRepnVisitor
from pyomo.repn.tests.test_linear import VisitorConfig
from pyomo.repn.util import InvalidNumber


class TestParameterizedLinearRepnVisitor(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 45))
        m.y = Var(domain=Binary)
        m.z = Var()

        return m

    def test_walk_sum(self):
        m = self.make_model()
        e = m.x + m.y
        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        self.assertEqual(repn.linear[id(m.x)], 1)
        self.assertIs(repn.constant, m.y)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x + m.y)

    def test_walk_triple_sum(self):
        m = self.make_model()
        e = m.x + m.z * m.y + m.z

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 2)
        self.assertIn(id(m.x), repn.linear)
        self.assertIn(id(m.y), repn.linear)
        self.assertEqual(repn.linear[id(m.x)], 1)
        self.assertIs(repn.linear[id(m.y)], m.z)
        self.assertIs(repn.constant, m.z)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x + m.z * m.y + m.z)

    def test_sum_two_of_the_same(self):
        # This hits the mult == 1 and vid in dest_dict case in _merge_dict
        m = self.make_model()
        e = m.x + m.x
        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        self.assertEqual(repn.linear[id(m.x)], 2)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), 2 * m.x)

    def test_sum_with_mult_0(self):
        m = self.make_model()
        e = 0 * m.x + m.x - m.y

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)
        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        self.assertEqual(repn.linear[id(m.x)], 1)
        assertExpressionsEqual(self, repn.constant, -m.y)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x - m.y)

    def test_sum_nonlinear_to_linear(self):
        m = self.make_model()
        e = m.y * m.x**2 + m.y * m.x - 3

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)
        assertExpressionsEqual(self, repn.nonlinear, m.y * m.x**2)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        self.assertIs(repn.linear[id(m.x)], m.y)
        self.assertEqual(repn.constant, -3)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), m.y * m.x - 3 + m.y * m.x**2
        )

    def test_sum_nonlinear_to_nonlinear(self):
        m = self.make_model()
        e = m.x**3 + 3 + m.x**2

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)
        assertExpressionsEqual(self, repn.nonlinear, m.x**3 + m.x**2)
        self.assertEqual(repn.constant, 3)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x**3 + m.x**2 + 3)

    def test_sum_to_linear_expr(self):
        m = self.make_model()
        e = m.x + m.y * (m.x + 5)

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.x)], 1 + m.y)
        assertExpressionsEqual(self, repn.constant, m.y * 5)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), (1 + m.y) * m.x + m.y * 5
        )

    def test_bilinear_term(self):
        m = self.make_model()
        e = m.x * m.y
        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        self.assertIs(repn.linear[id(m.x)], m.y)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.y * m.x)

    def test_distributed_bilinear_term(self):
        m = self.make_model()
        e = m.y * (m.x + 7)
        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.x), repn.linear)
        self.assertIs(repn.linear[id(m.x)], m.y)
        assertExpressionsEqual(self, repn.constant, m.y * 7)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.y * m.x + m.y * 7)

    def test_monomial(self):
        m = self.make_model()
        e = 45 * m.y
        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.x, m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.y), repn.linear)
        self.assertEqual(repn.linear[id(m.y)], 45)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), 45 * m.y)

    def test_constant(self):
        m = self.make_model()
        e = 45 * m.y
        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 0)
        assertExpressionsEqual(self, repn.constant, 45 * m.y)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), 45 * m.y)

    def test_fixed_var(self):
        m = self.make_model()
        m.x.fix(42)
        e = (m.y**2) * (m.x + m.x**2)

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)

        self.assertIsNone(repn.nonlinear)
        self.assertEqual(len(repn.linear), 0)
        assertExpressionsEqual(self, repn.constant, (m.y**2) * 1806)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.to_expression(visitor), (m.y**2) * 1806)

    def test_nonlinear(self):
        m = self.make_model()
        e = (m.y * log(m.x)) * (m.y + 2) / m.x

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z])

        repn = visitor.walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.nonlinear, log(m.x) * (m.y * (m.y + 2)) / m.x)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), log(m.x) * (m.y * (m.y + 2)) / m.x
        )

    def test_finalize(self):
        m = self.make_model()
        m.w = Var()

        e = m.x + 2 * m.w**2 * m.y - m.x - m.w * m.z

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.w]).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1, id(m.z): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 2)
        self.assertIn(id(m.y), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.y)], 2 * m.w**2)
        self.assertIn(id(m.z), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.z)], -m.w)
        self.assertEqual(repn.nonlinear, None)

        e *= 5

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.w]).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1, id(m.z): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 2)
        self.assertIn(id(m.y), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.y)], 5 * (2 * m.w**2))
        self.assertIn(id(m.z), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.z)], -5 * m.w)
        self.assertEqual(repn.nonlinear, None)

        e = 5 * (m.w * m.y + m.z**2 + 3 * m.w * m.y**3)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.w]).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.y): 0, id(m.z): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.y), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.y)], 5 * m.w)
        assertExpressionsEqual(self, repn.nonlinear, (m.z**2 + 3 * m.w * m.y**3) * 5)

    def test_ANY_over_constant_division(self):
        m = ConcreteModel()
        m.p = Param(mutable=True, initialize=2, domain=Any)
        m.x = Var()
        m.z = Var()
        m.y = Var()
        m.y.fix(1)

        expr = m.y + m.x + m.z + ((3 * m.z * m.x) / m.p) / m.y
        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.z]).walk_expression(expr)

        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 1 + m.z)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], 1 + 1.5 * m.z)
        self.assertEqual(repn.nonlinear, None)

    def test_errors_propagate_nan(self):
        m = ConcreteModel()
        m.p = Param(mutable=True, initialize=0, domain=Any)
        m.x = Var()
        m.z = Var()
        m.y = Var()
        m.y.fix(1)

        expr = m.y + m.x + m.z + ((3 * m.z * m.x) / m.p) / m.y
        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = ParameterizedLinearRepnVisitor(
                **cfg, wrt=[m.y, m.z]
            ).walk_expression(expr)
        self.assertEqual(
            LOG.getvalue(),
            "Exception encountered evaluating expression 'div(3*z, 0)'\n"
            "\tmessage: division by zero\n"
            "\texpression: 3*z*x/p\n",
        )
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, m.y + m.z)
        self.assertEqual(len(repn.linear), 1)
        self.assertIsInstance(repn.linear[id(m.x)], InvalidNumber)
        assertExpressionsEqual(self, repn.linear[id(m.x)].value, 1 + float('nan') / m.y)
        self.assertEqual(repn.nonlinear, None)

        m.y.fix(None)
        expr = m.z * log(m.y) + 3
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.z]).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertIsInstance(repn.constant, InvalidNumber)
        assertExpressionsEqual(self, repn.constant.value, float('nan') * m.z + 3)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_negation_constant(self):
        m = self.make_model()
        e = -(m.y * m.z + 17)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y, m.z]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, -1 * (m.y * m.z + 17))
        self.assertIsNone(repn.nonlinear)

    def test_product_nonlinear(self):
        m = self.make_model()
        e = (m.x**2) * (log(m.y) * m.z**4) * m.y
        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        assertExpressionsEqual(
            self, repn.nonlinear, (m.x**2) * (m.z**4 * log(m.y)) * m.y
        )

    def test_division_pseudo_constant_constant(self):
        m = self.make_model()
        e = m.x / 4 + m.y

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.x]).walk_expression(e)

        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.y), repn.linear)
        self.assertEqual(repn.linear[id(m.y)], 1)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, m.x / 4)
        self.assertIsNone(repn.nonlinear)

        e = 4 / m.x + m.y
        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.x]).walk_expression(e)

        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.y), repn.linear)
        self.assertEqual(repn.linear[id(m.y)], 1)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 4 / m.x)
        self.assertIsNone(repn.nonlinear)

        e = m.z / m.x + m.y
        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.x, m.z]).walk_expression(e)

        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.y), repn.linear)
        self.assertEqual(repn.linear[id(m.y)], 1)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, m.z / m.x)
        self.assertIsNone(repn.nonlinear)

    def test_division_ANY_pseudo_constant(self):
        m = self.make_model()
        e = (m.x + 3 * m.z) / m.y

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y]).walk_expression(e)

        self.assertEqual(len(repn.linear), 2)
        self.assertIn(id(m.x), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.x)], 1 / m.y)
        self.assertIn(id(m.z), repn.linear)
        assertExpressionsEqual(self, repn.linear[id(m.z)], (1 / m.y) * 3)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertIsNone(repn.nonlinear)

    def test_duplicate(self):
        m = self.make_model()
        e = (1 + m.x) ** 2 + m.y

        cfg = VisitorConfig()
        visitor = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y])
        visitor.max_exponential_expansion = 2
        repn = visitor.walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertIs(repn.constant, m.y)
        assertExpressionsEqual(self, repn.nonlinear, (m.x + 1) * (m.x + 1))

    def test_pow_ANY_pseudo_constant(self):
        m = self.make_model()
        e = (m.x**2 + 3 * m.z) ** m.y

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        assertExpressionsEqual(self, repn.nonlinear, (m.x**2 + 3 * m.z) ** m.y)

    def test_pow_pseudo_constant_ANY(self):
        m = self.make_model()
        e = m.y ** (m.x**2 + 3 * m.z)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        assertExpressionsEqual(self, repn.nonlinear, m.y ** (m.x**2 + 3 * m.z))

    def test_pow_linear_pseudo_constant(self):
        m = self.make_model()
        e = (m.x + 3 * m.z) ** m.y

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        assertExpressionsEqual(self, repn.nonlinear, (m.x + 3 * m.z) ** m.y)

    def test_pow_pseudo_constant_linear(self):
        m = self.make_model()
        e = m.y ** (m.x + 3 * m.z)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.y]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        assertExpressionsEqual(self, repn.nonlinear, m.y ** (m.x + 3 * m.z))

    def test_0_mult(self):
        m = self.make_model()
        m.p = Var()
        m.p.fix(0)
        e = m.p * (m.y**2 + m.z)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.z]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.nonlinear, (m.y**2) * 0)
        assertExpressionsEqual(self, repn.constant, 0 * m.z)

    def test_0_mult_nan(self):
        m = self.make_model()
        m.p = Param(initialize=0, mutable=True)
        m.y.domain = Any
        m.y.fix(float('nan'))
        e = m.p * (m.y**2 + m.x)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.x]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertIsNone(repn.nonlinear)
        self.assertIsInstance(repn.constant, InvalidNumber)
        assertExpressionsEqual(self, repn.constant.value, 0 * (float('nan') + m.x))

    def test_0_mult_nan_param(self):
        m = self.make_model()
        m.p = Param(initialize=0, mutable=True)
        m.y.fix(float('nan'))
        e = m.p * (m.y**2)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[]).walk_expression(e)

        self.assertEqual(len(repn.linear), 0)
        self.assertEqual(repn.multiplier, 1)
        self.assertIsNone(repn.nonlinear)
        self.assertIsInstance(repn.constant, InvalidNumber)
        assertExpressionsEqual(self, repn.constant.value, 0 * float('nan'))

    def test_0_mult_linear_with_nan(self):
        m = self.make_model()
        m.p = Param(initialize=0, mutable=True)
        m.x.domain = Any
        m.x.fix(float('nan'))
        e = m.p * (3 * m.x * m.y + m.z)

        cfg = VisitorConfig()
        repn = ParameterizedLinearRepnVisitor(**cfg, wrt=[m.z]).walk_expression(e)
        self.assertEqual(len(repn.linear), 1)
        self.assertIn(id(m.y), repn.linear)
        self.assertIsInstance(repn.linear[id(m.y)], InvalidNumber)
        assertExpressionsEqual(self, repn.linear[id(m.y)].value, 0 * 3 * float('nan'))
        self.assertEqual(repn.multiplier, 1)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.constant, 0 * m.z)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os

# import os.path
import subprocess

from pyomo.common import Executable
from pyomo.common.collections import Bunch
from pyomo.common.tempfiles import TempfileManager

from pyomo.opt.base import ProblemFormat, ResultsFormat
from pyomo.opt.base.solvers import _extract_version, SolverFactory
from pyomo.opt.results import SolverStatus, TerminationCondition, SolutionStatus
from pyomo.opt.solver import SystemCallSolver

import logging

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('scip', doc='The SCIP LP/MIP solver')
class SCIPAMPL(SystemCallSolver):
    """A generic optimizer that uses the AMPL Solver Library to interface with applications."""

    # Cache default executable, so we do not need to repeatedly query the
    # versions every time.
    _known_versions = {}

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds["type"] = "scip"
        SystemCallSolver.__init__(self, **kwds)
        #
        # Setup valid problem formats, and valid results for each problem format
        # Also set the default problem and results formats.
        #
        self._valid_problem_formats = [ProblemFormat.nl]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.nl] = [ResultsFormat.sol]
        self.set_problem_format(ProblemFormat.nl)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.integer = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def _default_results_format(self, prob_format):
        return ResultsFormat.sol

    def _default_executable(self):
        executable = Executable("scip")

        if executable:
            executable_path = executable.path()
            if executable_path not in self._known_versions:
                self._known_versions[executable_path] = self._get_version(
                    executable_path
                )
            _ver = self._known_versions[executable_path]
            if _ver and _ver >= (8,):
                return executable_path

        # revert to scipampl for older versions
        executable = Executable("scipampl")
        if not executable:
            logger.warning(
                "Could not locate the 'scip' executable or"
                " the older 'scipampl' executable, which is "
                "required for solver %s" % self.name
            )
            self.enable = False
            return None
        return executable.path()

    def _get_version(self, solver_exec=None):
        """
        Returns a tuple describing the solver executable version.
        """
        if solver_exec is None:
            solver_exec = self.executable()
            if solver_exec is None:
                return _extract_version('')
        results = subprocess.run(
            [solver_exec, "--version"],
            timeout=self._version_timeout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        return _extract_version(results.stdout)

    def create_command_line(self, executable, problem_files):
        assert self._problem_format == ProblemFormat.nl
        assert self._results_format == ResultsFormat.sol

        #
        # Define log file
        #
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix="_scip.log")

        fname = problem_files[0]
        if '.' in fname:
            tmp = fname.split('.')
            if len(tmp) > 2:
                fname = '.'.join(tmp[:-1])
            else:
                fname = tmp[0]
        self._soln_file = fname + ".sol"

        #
        # Define results file (since an external parser is used)
        #
        self._results_file = self._soln_file

        #
        # Define command line
        #
        env = os.environ.copy()
        #
        # Merge the PYOMO_AMPLFUNC (externals defined within
        # Pyomo/Pyomo) with any user-specified external function
        # libraries
        #
        if 'PYOMO_AMPLFUNC' in env:
            if 'AMPLFUNC' in env:
                env['AMPLFUNC'] += "\n" + env['PYOMO_AMPLFUNC']
            else:
                env['AMPLFUNC'] = env['PYOMO_AMPLFUNC']

        # Since Version 8.0.0 .nl problem file paths should be provided without the .nl
        # extension
        if executable not in self._known_versions:
            self._known_versions[executable] = self._get_version(executable)
        _ver = self._known_versions[executable]
        if _ver and _ver >= (8, 0, 0):
            problem_file = os.path.splitext(problem_files[0])[0]
        else:
            problem_file = problem_files[0]

        cmd = [executable, problem_file, '-AMPL']
        if self._timer:
            cmd.insert(0, self._timer)

        # GAH: I am going to re-add the code by Zev that passed options through
        # to the command line. I'm not sure what solvers this method of passing options
        # through the envstr variable works for, but it does not seem to work for cplex
        # or gurobi
        env_opt = []
        of_opt = []
        for key in self.options:
            if key == 'solver':
                continue
            if isinstance(self.options[key], str) and ' ' in self.options[key]:
                env_opt.append(key + "=\"" + str(self.options[key]) + "\"")
            else:
                env_opt.append(key + "=" + str(self.options[key]))
            of_opt.append(str(key) + " = " + str(self.options[key]))

        if (
            self._timelimit is not None
            and self._timelimit > 0.0
            and 'limits/time' not in self.options
        ):
            of_opt.append("limits/time = " + str(self._timelimit))

        envstr = "%s_options" % self.options.solver
        # Merge with any options coming in through the environment
        env[envstr] = " ".join(env_opt)

        if len(of_opt) > 0:
            # Now check if an 'scip.set' file exists in the
            # current working directory. If so, we need to
            # make it clear that this file will be ignored.
            default_of_name = os.path.join(os.getcwd(), 'scip.set')
            if os.path.exists(default_of_name):
                logger.warning(
                    "A file named '%s' exists in "
                    "the current working directory, but "
                    "SCIP options are being set using a "
                    "separate options file. The options "
                    "file '%s' will be ignored." % (default_of_name, default_of_name)
                )

            options_dir = TempfileManager.create_tempdir()
            # Now write the new options file
            with open(os.path.join(options_dir, 'scip.set'), 'w') as f:
                for line in of_opt:
                    f.write(line + "\n")
        else:
            options_dir = None

        return Bunch(cmd=cmd, log_file=self._log_file, env=env, cwd=options_dir)

    def _postsolve(self):
        # find SCIP version (calling version() and _get_version() mess things)

        executable = self._command.cmd[0]

        version = self._known_versions[executable]

        if version < (8, 0, 0, 0):
            # it may be possible to get results from older version but this was
            # not tested, so the old way of doing things is here preserved

            results = super(SCIPAMPL, self)._postsolve()

        else:
            # repeat code from super(SCIPAMPL, self)._postsolve()
            # in order to access the log file and get the results from there

            if self._log_file is not None:
                OUTPUT = open(self._log_file, "w")
                OUTPUT.write("Solver command line: " + str(self._command.cmd) + '\n')
                OUTPUT.write("\n")
                OUTPUT.write(self._log + '\n')
                OUTPUT.close()

            # JPW: The cleanup of the problem file probably shouldn't be here, but
            #   rather in the base OptSolver class. That would require movement of
            #   the keepfiles attribute and associated cleanup logic to the base
            #   class, which I didn't feel like doing at this present time. the
            #   base class remove_files method should clean up the problem file.

            if (self._log_file is not None) and (not os.path.exists(self._log_file)):
                msg = "File '%s' not generated while executing %s"
                raise IOError(msg % (self._log_file, self.path))
            results = None

            if self._results_format is not None:
                results = self.process_output(self._rc)

                # read results from the log file

                log_dict = self.read_scip_log(self._log_file)

                if len(log_dict) != 0:
                    # if any were read, store them

                    results.solver.time = log_dict['solving_time']
                    results.solver.gap = log_dict['gap']
                    results.solver.node_count = log_dict['solving_nodes']
                    results.solver.primal_bound = log_dict['primal_bound']
                    results.solver.dual_bound = log_dict['dual_bound']

                # TODO: get scip to produce a statistics file and read it
                # Why? It has all the information one can possibly need.
                #
                # If keepfiles is true, then we pop the
                # TempfileManager context while telling it to
                # _not_ remove the files.
                #
                if not self._keepfiles:
                    # in some cases, the solution filename is
                    # not generated via the temp-file mechanism,
                    # instead being automatically derived from
                    # the input lp/nl filename. so, we may have
                    # to clean it up manually.
                    if (not self._soln_file is None) and os.path.exists(
                        self._soln_file
                    ):
                        os.remove(self._soln_file)

            TempfileManager.pop(remove=not self._keepfiles)

        # **********************************************************************
        # **********************************************************************

        # UNKNOWN # unknown='unknown' # An uninitialized value

        if "unknown" in results.solver.message:
            results.solver.status = SolverStatus.unknown
            results.solver.termination_condition = TerminationCondition.unknown
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.unknown

        # ABORTED # userInterrupt='userInterrupt' # Interrupt signal generated by user

        elif "user interrupt" in results.solver.message:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_condition = TerminationCondition.userInterrupt
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.unknown

        # OK # maxEvaluations='maxEvaluations' # Exceeded maximum number of problem evaluations

        elif "node limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxEvaluations
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # maxEvaluations='maxEvaluations' # Exceeded maximum number of problem evaluations

        elif "total node limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxEvaluations
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # maxEvaluations='maxEvaluations' # Exceeded maximum number of problem evaluations

        elif "stall node limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxEvaluations
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # maxTimeLimit='maxTimeLimit' # Exceeded maximum time limited allowed by user but having return a feasible solution

        elif "time limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxTimeLimit
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # other='other' # Other, uncategorized normal termination

        elif "memory limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.other
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # other='other' # Other, uncategorized normal termination

        elif "gap limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.other
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # other='other' # Other, uncategorized normal termination

        elif "solution limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.other
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # other='other' # Other, uncategorized normal termination

        elif "solution improvement limit reached" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.other
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.stoppedByLimit

        # OK # optimal='optimal' # Found an optimal solution

        elif "optimal solution" in results.solver.message:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.optimal
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.optimal
            try:
                if results.solver.primal_bound < results.solver.dual_bound:
                    results.problem.lower_bound = results.solver.primal_bound
                    results.problem.upper_bound = results.solver.dual_bound
                else:
                    results.problem.lower_bound = results.solver.dual_bound
                    results.problem.upper_bound = results.solver.primal_bound
            except AttributeError:
                """
                This may occur if SCIP solves the problem during presolve. In that case,
                the log file may not get parsed correctly (self.read_scip_log), and
                results.solver.primal_bound will not be populated.
                """
                pass

        # WARNING # infeasible='infeasible' # Demonstrated that the problem is infeasible

        elif "infeasible" in results.solver.message:
            results.solver.status = SolverStatus.warning
            results.solver.termination_condition = TerminationCondition.infeasible
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.infeasible

        # WARNING # unbounded='unbounded' # Demonstrated that problem is unbounded

        elif "unbounded" in results.solver.message:
            results.solver.status = SolverStatus.warning
            results.solver.termination_condition = TerminationCondition.unbounded
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.unbounded

        # WARNING # infeasibleOrUnbounded='infeasibleOrUnbounded'   # Problem is either infeasible or unbounded

        elif "infeasible or unbounded" in results.solver.message:
            results.solver.status = SolverStatus.warning
            results.solver.termination_condition = (
                TerminationCondition.infeasibleOrUnbounded
            )
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.unsure

        # UNKNOWN # unknown='unknown' # An uninitialized value

        else:
            logger.warning(
                "Unexpected SCIP solver message: %s" % (results.solver.message)
            )
            results.solver.status = SolverStatus.unknown
            results.solver.termination_condition = TerminationCondition.unknown
            if len(results.solution) > 0:
                results.solution(0).status = SolutionStatus.unknown

        return results

    @staticmethod
    def read_scip_log(filename: str):
        # TODO: check file exists, ensure opt has finished, etc

        from collections import deque

        with open(filename) as f:
            scip_lines = list(deque(f, 7))
            scip_lines.pop()

        expected_labels = [
            'SCIP Status        :',
            'Solving Time (sec) :',
            'Solving Nodes      :',
            'Primal Bound       :',
            'Dual Bound         :',
            'Gap                :',
        ]

        colon_position = 19  # or scip_lines[0].index(':')

        for i, log_file_line in enumerate(scip_lines):
            if expected_labels[i] != log_file_line[0 : colon_position + 1]:
                return {}

        # get data

        solver_status = scip_lines[0][colon_position + 2 : scip_lines[0].index('\n')]

        solving_time = float(
            scip_lines[1][colon_position + 2 : scip_lines[1].index('\n')].split(' ')[0]
        )

        try:
            solving_nodes = int(
                scip_lines[2][colon_position + 2 : scip_lines[2].index('(')]
            )
        except ValueError:
            solving_nodes = int(
                scip_lines[2][colon_position + 2 : scip_lines[2].index('\n')]
            )

        primal_bound = float(
            scip_lines[3][colon_position + 2 : scip_lines[3].index('(')]
        )

        dual_bound = float(
            scip_lines[4][colon_position + 2 : scip_lines[4].index('\n')]
        )

        try:
            gap = float(scip_lines[5][colon_position + 2 : scip_lines[5].index('%')])
        except ValueError:
            gap = scip_lines[5][colon_position + 2 : scip_lines[5].index('\n')]

            if gap == 'infinite':
                gap = float('inf')

        out_dict = {
            'solver_status': solver_status,
            'solving_time': solving_time,
            'solving_nodes': solving_nodes,
            'primal_bound': primal_bound,
            'dual_bound': dual_bound,
            'gap': gap,
        }

        return out_dict



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import os
import subprocess
import time
import datetime
from io import StringIO
import sys
import struct
import re
import pathlib

from pyomo.common.dependencies import attempt_import
from pyomo.common.fileutils import Executable, ExecutableData
from pyomo.common.config import (
    ConfigValue,
    ConfigDict,
    document_configdict,
    Path,
    document_class_CONFIG,
)
from pyomo.common.tempfiles import TempfileManager
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.base import value, Objective
from pyomo.core.staleflag import StaleFlagManager
from pyomo.contrib.solver.common.base import SolverBase, Availability
from pyomo.contrib.solver.common.config import SolverConfig
from pyomo.contrib.solver.common.results import (
    Results,
    SolutionStatus,
    TerminationCondition,
)
from pyomo.contrib.solver.solvers.gms_sol_reader import GMSSolutionLoader

import pyomo.core.base.suffix
from pyomo.common.tee import TeeStream
from pyomo.core.expr.visitor import replace_expressions
from pyomo.core.base.suffix import Suffix
from pyomo.common.errors import ApplicationError
from pyomo.contrib.solver.common.util import NoOptimalSolutionError
from pyomo.repn.plugins.gams_writer_v2 import GAMSWriter

logger = logging.getLogger(__name__)


def _gams_importer():
    try:
        import gams.core.gdx as gdxcc
    except ImportError:
        try:
            # fall back to the pre-GAMS-45.0 API
            import gdxcc
        except:
            # suppress the error from the old API and reraise the
            # current API import error
            pass
        raise
    return gdxcc


gdxcc, gdxcc_available = attempt_import('gdxcc', importer=_gams_importer)


@document_configdict()
class GAMSConfig(SolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )
        self.executable: ExecutableData = self.declare(
            'executable',
            ConfigValue(
                domain=Executable,
                default='gams',
                description="Executable for gams. Defaults to searching the "
                "``PATH`` for the first available ``gams``.",
            ),
        )
        self.logfile: str = self.declare(
            'logfile',
            ConfigValue(
                domain=Path(),
                default=None,
                description="Filename to output GAMS log to a file.",
            ),
        )
        self.writer_config: ConfigDict = self.declare(
            'writer_config', GAMSWriter.CONFIG()
        )


@document_class_CONFIG(methods=['solve'])
class GAMS(SolverBase):
    #: Global class configuration;
    #: see :ref:`pyomo.contrib.solver.solvers.gams.GAMS::CONFIG`.
    CONFIG = GAMSConfig()

    # default behaviour of gams is to print to console, for
    # compatibility with windows and *nix we want to explicitly log to
    # stdout (see https://www.gams.com/latest/docs/UG_GamsCall.html)
    _log_levels = {
        (True, False): "lo=3",
        (False, False): "lo=0",
        (False, True): "lo=2",
        (True, True): "lo=4",
    }

    #: cache of availability / version information
    _exe_cache: dict = {None: (None, Availability.NotFound)}

    # mapping for GAMS SOLVESTAT
    _SOLVER_STATUS_LOOKUP = {
        1: None,
        2: TerminationCondition.iterationLimit,
        3: TerminationCondition.maxTimeLimit,
        4: TerminationCondition.error,
        5: TerminationCondition.iterationLimit,
        6: TerminationCondition.error,
        7: TerminationCondition.licensingProblems,
        8: TerminationCondition.interrupted,
        9: TerminationCondition.error,
        10: TerminationCondition.error,
        11: TerminationCondition.error,
        12: TerminationCondition.error,
        13: TerminationCondition.error,
    }

    # mapping for GAMS MODELSTAT
    _MODEL_STATUS_LOOKUP = {
        1: (SolutionStatus.optimal, TerminationCondition.convergenceCriteriaSatisfied),
        2: (SolutionStatus.feasible, TerminationCondition.convergenceCriteriaSatisfied),
        3: (SolutionStatus.noSolution, TerminationCondition.unbounded),
        4: (SolutionStatus.infeasible, TerminationCondition.provenInfeasible),
        5: (SolutionStatus.infeasible, TerminationCondition.locallyInfeasible),
        6: (SolutionStatus.infeasible, TerminationCondition.unknown),
        7: (SolutionStatus.feasible, TerminationCondition.unknown),
        8: (SolutionStatus.feasible, TerminationCondition.unknown),
        9: (SolutionStatus.noSolution, TerminationCondition.unknown),
        10: (SolutionStatus.infeasible, TerminationCondition.infeasibleOrUnbounded),
        11: (SolutionStatus.noSolution, TerminationCondition.licensingProblems),
        12: (SolutionStatus.noSolution, TerminationCondition.unknown),
        13: (SolutionStatus.noSolution, TerminationCondition.error),
        14: (SolutionStatus.noSolution, TerminationCondition.error),
        15: (SolutionStatus.optimal, TerminationCondition.convergenceCriteriaSatisfied),
        16: (
            SolutionStatus.feasible,
            TerminationCondition.convergenceCriteriaSatisfied,
        ),
        17: (
            SolutionStatus.feasible,
            TerminationCondition.convergenceCriteriaSatisfied,
        ),
        18: (SolutionStatus.noSolution, TerminationCondition.unbounded),
        19: (SolutionStatus.infeasible, TerminationCondition.provenInfeasible),
    }

    def __init__(self, **kwds):
        super().__init__(**kwds)

        #: Instance configuration;
        #: see :ref:`pyomo.contrib.solver.solvers.gams.GAMS::CONFIG`.
        self.config = self.config

    def available(self) -> Availability:
        ver, avail = self._get_version(self.config.executable.path())
        return avail

    def version(self) -> tuple[int, int, int] | None:
        ver, avail = self._get_version(self.config.executable.path())
        return ver

    def _get_version(self, exe: str | None):
        # check the cache
        if exe in self._exe_cache:
            return self._exe_cache[exe]

        # Note: non-None str paths are guaranteed to exist and be executable files
        res = subprocess.run(
            [exe],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf8",
            check=False,
        )
        if res.returncode:
            params = (None, Availability.NotFound)
            self._exe_cache[exe] = params
            out = res.stdout.replace('\n', '\n    ').strip()
            logger.warning(
                "Failed running GAMS command to get version (non-zero returncode "
                f"{res.returncode}):\n    {out}"
            )
            return params

        version_pattern = r"GAMS Release *: *(\d{1,3}\.\d{1,3}\.\d{1,3})"
        found = re.search(version_pattern, res.stdout)

        if not found:
            params = (None, Availability.NotFound)
            self._exe_cache[exe] = params
            out = res.stdout.replace('\n', '\n    ').strip()
            logger.warning(
                "Failed parsing GAMS version (version not found while parsing):"
                f"\n    {out}"
            )
            return params

        version = found.group(1)
        version = tuple(int(i) for i in version.split('.'))

        # TBD: does this also catch Community licenses?
        if "GAMS Demo" in res.stdout:
            avail = Availability.LimitedLicense
        else:
            avail = Availability.FullLicense

        # TBD: should we run a small problem (1-variable LP) to make
        # sure the license is still valid?  Alternatively, we can leave
        # Availability as NOTSET until someone explicitly *asks* for
        # .available(), in which case we can justify running the small
        # model (potentially requiring us to check out a license from a
        # license server).

        params = (version, avail)
        self._exe_cache[exe] = params
        return params

    def solve(self, model, **kwds):
        ####################################################################
        # Presolve
        ####################################################################
        # Begin time tracking
        start_timestamp = datetime.datetime.now(datetime.timezone.utc)
        tick = time.perf_counter()

        # Update configuration options, based on keywords passed to solve
        # preserve_implicit=True is required to extract solver_options ConfigDict
        config: GAMSConfig = self.config(value=kwds, preserve_implicit=True)

        # Check if the solver executable exists:
        if not config.executable:
            raise ApplicationError(
                f"{self.__class__.__name__}: 'gams' executable not found"
            )

        if config.timer is None:
            config.timer = HierarchicalTimer()
        timer = config.timer
        StaleFlagManager.mark_all_as_stale()

        # update the writer config if any of the overlapping
        # keys exists in the solver_options

        for key, val in config.solver_options.items():
            config.writer_config.gams_commands.append(f"option {key}={val};")

        if not config.writer_config.put_results_format:
            config.writer_config.put_results_format = (
                'gdx' if gdxcc_available else 'dat'
            )

        # Copy resLim last so that time_limit overrides any limit set in
        # solver_options (or writer_config.gams_commands)
        if config.time_limit is not None:
            config.writer_config.gams_commands.append(
                f"option resLim={config.time_limit};"
            )

        # local variable to hold the working directory name and flags
        dname = None
        lst = "output.lst"
        model_name = "model"

        output_filename = None
        with TempfileManager.new_context() as tempfile:
            # IMPORTANT - only delete the whole tmpdir if the solver was the one
            # that made the directory. Otherwise, just delete the files the solver
            # made, if not keepfiles. That way the user can select a directory
            # they already have, like the current directory, without having to
            # worry about the rest of the contents of that directory being deleted.
            if not config.working_dir:
                dname = tempfile.mkdtemp()
            else:
                dname = config.working_dir
            if not os.path.exists(dname):
                os.mkdir(dname)
            basename = os.path.join(dname, model_name)
            output_filename = basename + '.gms'
            lst_filename = os.path.join(dname, lst)

            timer.start(f'write_gms_file')
            with open(output_filename, 'w', newline='\n', encoding='utf-8') as gms_file:
                gms_info = GAMSWriter().write(
                    model, gms_file, config=config.writer_config
                )
                # NOTE: omit InfeasibleConstraintException for now
            timer.stop(f'write_gms_file')

            if config.writer_config.put_results_format == 'gdx':
                results_filename = os.path.join(dname, "GAMS_MODEL_p.gdx")
                statresults_filename = os.path.join(
                    dname, "%s_s.gdx" % (config.writer_config.put_results,)
                )
            else:
                results_filename = os.path.join(
                    dname, "%s.dat" % (config.writer_config.put_results,)
                )
                statresults_filename = os.path.join(
                    dname, "%sstat.dat" % (config.writer_config.put_results,)
                )

            ####################################################################
            # Apply solver
            ####################################################################
            exe_path = config.executable.path()
            command = [exe_path, output_filename, "o=" + lst, "curdir=" + dname]

            # handled tee and logfile based on the length of list and
            # string respectively
            command.append(self._log_levels[(bool(config.tee), bool(config.logfile))])

            ostreams = [StringIO()]
            if config.tee:
                ostreams.append(sys.stdout)

            with TeeStream(*ostreams) as t:
                timer.start('subprocess')
                subprocess_result = subprocess.run(
                    command, stdout=t.STDOUT, stderr=t.STDERR, cwd=dname
                )
                timer.stop('subprocess')
            rc = subprocess_result.returncode
            txt = ostreams[0].getvalue()
            if config.working_dir:
                logger.info("\nGAMS WORKING DIRECTORY: %s\n" % config.working_dir)

            if rc:
                # If nothing was raised, or for all other cases, raise this
                error_message = f"GAMS process encountered an error (returncode={rc})."
                if rc == 3:
                    # Execution Error
                    # Run check_expr_evaluation, which errors if necessary
                    error_message += (
                        "\nError rc=3 (GAMS execution error), to be determined later."
                    )
                error_message += "\nCheck listing file for details.\n"
                logger.error(error_message)
                logger.error(txt.strip())
                if os.path.exists(lst_filename):
                    with open(lst_filename, 'r') as FILE:
                        logger.error(
                            "\nGAMS Listing file:\n\n%s" % (FILE.read().strip(),)
                        )
                raise RuntimeError(error_message)

            timer.start('parse_results')
            if config.writer_config.put_results_format == 'gdx':
                model_soln, stat_vars = self._parse_gdx_results(
                    config, results_filename, statresults_filename
                )
            else:
                model_soln, stat_vars = self._parse_dat_results(
                    config, results_filename, statresults_filename
                )
            timer.stop('parse_results')

            ####################################################################
            # Postsolve (WIP)
            results = self._postsolve(
                model, timer, config, model_soln, stat_vars, gms_info
            )

            results.solver_config = config
            results.solver_log = ostreams[0].getvalue()

            tock = time.perf_counter()
            results.timing_info.start_timestamp = start_timestamp
            results.timing_info.wall_time = tock - tick
            results.timing_info.timer = timer
            return results

    def _postsolve(self, model, timer, config, model_soln, stat_vars, gms_info):
        model_suffixes = list(
            name
            for (name, comp) in pyomo.core.base.suffix.active_import_suffix_generator(
                model
            )
        )
        extract_dual = 'dual' in model_suffixes
        extract_rc = 'rc' in model_suffixes
        results = Results()
        results.solver_name = "GAMS "
        results.solver_version = self.version()

        # --- Process GAMS SOLVESTAT ---
        solvestat = stat_vars["SOLVESTAT"]
        solver_term = GAMS._SOLVER_STATUS_LOOKUP.get(
            solvestat, TerminationCondition.unknown
        )

        # specific message for status 4
        if solvestat == 4:
            results.message = "Solver quit with a problem (see LST file)"

        # --- Process GAMS MODELSTAT ---
        modelstat = stat_vars["MODELSTAT"]
        solution_status, model_term = GAMS._MODEL_STATUS_LOOKUP.get(
            modelstat, (SolutionStatus.noSolution, TerminationCondition.unknown)
        )

        # --- Populate 'results'
        results.solution_status = solution_status

        # replaced below, if solution should be loaded
        results.solution_loader = GMSSolutionLoader(None, None)

        if solvestat == 1:
            results.termination_condition = model_term
        else:
            results.termination_condition = solver_term

        # extra_info (save GAMS status codes)
        results.extra_info.gams_solvestat = solvestat
        results.extra_info.gams_modelstat = modelstat

        # Taken from ipopt.py
        if (
            config.raise_exception_on_nonoptimal_result
            and results.solution_status != SolutionStatus.optimal
        ):
            raise NoOptimalSolutionError()

        obj = list(model.component_data_objects(Objective, active=True))

        if results.solution_status in {SolutionStatus.feasible, SolutionStatus.optimal}:
            results.solution_loader = GMSSolutionLoader(
                gdx_data=model_soln, gms_info=gms_info
            )

            if config.load_solutions:
                results.solution_loader.load_vars()
                if len(obj) == 1:
                    results.incumbent_objective = stat_vars["OBJVAL"]
                else:
                    results.incumbent_objective = None
                if (
                    hasattr(model, 'dual')
                    and isinstance(model.dual, Suffix)
                    and model.dual.import_enabled()
                ):
                    model.dual.update(results.solution_loader.get_duals())
                if (
                    hasattr(model, 'rc')
                    and isinstance(model.rc, Suffix)
                    and model.rc.import_enabled()
                ):
                    model.rc.update(results.solution_loader.get_reduced_costs())

            else:
                results.incumbent_objective = value(
                    replace_expressions(
                        obj[0].expr,
                        substitution_map={
                            id(v): val
                            for v, val in results.solution_loader.get_primals().items()
                        },
                        descend_into_named_expressions=True,
                        remove_named_expressions=True,
                    )
                )

        return results

    def _parse_gdx_results(self, config, results_filename, statresults_filename):
        model_soln = dict()
        stat_vars = dict.fromkeys(
            [
                'MODELSTAT',
                'SOLVESTAT',
                'OBJEST',
                'OBJVAL',
                'NUMVAR',
                'NUMEQU',
                'NUMDVAR',
                'NUMNZ',
                'ETSOLVE',
            ]
        )

        pgdx = gdxcc.new_gdxHandle_tp()
        ret = gdxcc.gdxCreateD(pgdx, os.path.dirname(config.executable.path()), 128)
        if not ret[0]:
            raise RuntimeError("GAMS GDX failure (gdxCreate): %s." % ret[1])
        if os.path.exists(statresults_filename):
            ret = gdxcc.gdxOpenRead(pgdx, statresults_filename)
            if not ret[0]:
                raise RuntimeError("GAMS GDX failure (gdxOpenRead): %d." % ret[1])

            specVals = gdxcc.doubleArray(gdxcc.GMS_SVIDX_MAX)
            rc = gdxcc.gdxGetSpecialValues(pgdx, specVals)

            specVals[gdxcc.GMS_SVIDX_EPS] = sys.float_info.min
            specVals[gdxcc.GMS_SVIDX_UNDEF] = float("nan")
            specVals[gdxcc.GMS_SVIDX_PINF] = float("inf")
            specVals[gdxcc.GMS_SVIDX_MINF] = float("-inf")
            specVals[gdxcc.GMS_SVIDX_NA] = struct.unpack(
                ">d", bytes.fromhex("fffffffffffffffe")
            )[0]
            gdxcc.gdxSetSpecialValues(pgdx, specVals)

            i = 0
            while True:
                i += 1
                ret = gdxcc.gdxDataReadRawStart(pgdx, i)
                if not ret[0]:
                    break

                ret = gdxcc.gdxSymbolInfo(pgdx, i)
                if not ret[0]:
                    break
                if len(ret) < 2:
                    raise RuntimeError("GAMS GDX failure (gdxSymbolInfo).")
                stat = ret[1]
                if not stat in stat_vars:
                    continue

                ret = gdxcc.gdxDataReadRaw(pgdx)
                if not ret[0] or len(ret[2]) == 0:
                    raise RuntimeError("GAMS GDX failure (gdxDataReadRaw).")

                if stat in ('OBJEST', 'OBJVAL', 'ETSOLVE'):
                    stat_vars[stat] = ret[2][0]
                else:
                    stat_vars[stat] = int(ret[2][0])

            gdxcc.gdxDataReadDone(pgdx)
            gdxcc.gdxClose(pgdx)

        if os.path.exists(results_filename):
            ret = gdxcc.gdxOpenRead(pgdx, results_filename)
            if not ret[0]:
                raise RuntimeError("GAMS GDX failure (gdxOpenRead): %d." % ret[1])

            specVals = gdxcc.doubleArray(gdxcc.GMS_SVIDX_MAX)
            rc = gdxcc.gdxGetSpecialValues(pgdx, specVals)

            specVals[gdxcc.GMS_SVIDX_EPS] = sys.float_info.min
            specVals[gdxcc.GMS_SVIDX_UNDEF] = float("nan")
            specVals[gdxcc.GMS_SVIDX_PINF] = float("inf")
            specVals[gdxcc.GMS_SVIDX_MINF] = float("-inf")
            specVals[gdxcc.GMS_SVIDX_NA] = struct.unpack(
                ">d", bytes.fromhex("fffffffffffffffe")
            )[0]
            gdxcc.gdxSetSpecialValues(pgdx, specVals)

            i = 0
            while True:
                i += 1
                ret = gdxcc.gdxDataReadRawStart(pgdx, i)
                if not ret[0]:
                    break

                ret = gdxcc.gdxDataReadRaw(pgdx)
                if not ret[0] or len(ret[2]) < 2:
                    raise RuntimeError("GAMS GDX failure (gdxDataReadRaw).")
                level = ret[2][0]
                dual = ret[2][1]

                ret = gdxcc.gdxSymbolInfo(pgdx, i)
                if not ret[0]:
                    break
                if len(ret) < 2:
                    raise RuntimeError("GAMS GDX failure (gdxSymbolInfo).")
                model_soln[ret[1]] = (level, dual)

            gdxcc.gdxDataReadDone(pgdx)
            gdxcc.gdxClose(pgdx)

        gdxcc.gdxFree(pgdx)
        gdxcc.gdxLibraryUnload()
        return model_soln, stat_vars

    def _parse_dat_results(self, config, results_filename, statresults_filename):
        with open(statresults_filename, 'r') as statresults_file:
            statresults_text = statresults_file.read()

        stat_vars = dict()

        # Skip first line of explanatory text
        for line in statresults_text.splitlines()[1:]:
            items = line.split()

            try:
                numval = float(items[1])
            except:
                numval = float("nan")

            stat_vars[items[0]] = numval

        with open(results_filename, 'r') as results_file:
            results_text = results_file.read()

        model_soln = dict()
        # Skip first line of explanatory text
        for line in results_text.splitlines()[1:]:
            items = line.split()
            model_soln[items[0]] = (float(items[1]), float(items[2]))

        return model_soln, stat_vars



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.environ import (
    AbstractModel,
    Param,
    RangeSet,
    Var,
    Reals,
    Binary,
    PositiveIntegers,
    Constraint,
    Objective,
)
import math

model = AbstractModel()

model.N = Param(within=PositiveIntegers)

model.Locations = RangeSet(1, model.N)

model.P = Param(within=RangeSet(1, model.N))

model.M = Param(within=PositiveIntegers)

model.Customers = RangeSet(1, model.M)


def d_rule(model, n, m):
    return math.sin(n * 2.33333 + m * 7.99999)


model.d = Param(model.Locations, model.Customers, initialize=d_rule, within=Reals)

model.x = Var(model.Locations, model.Customers, bounds=(0.0, 1.0))

model.y = Var(model.Locations, within=Binary)


def rule(model):
    return sum(
        [
            model.d[n, m] * model.x[n, m]
            for n in model.Locations
            for m in model.Customers
        ]
    )


model.obj = Objective(rule=rule)


def rule(model, m):
    return (sum([model.x[n, m] for n in model.Locations]), 1.0)


model.single_x = Constraint(model.Customers, rule=rule)


def rule(model, n, m):
    return (None, model.x[n, m] - model.y[n], 0.0)


model.bound_y = Constraint(model.Locations, model.Customers, rule=rule)


def rule(model):
    return (sum([model.y[n] for n in model.Locations]) - model.P, 0.0)


model.num_facilities = Constraint(rule=rule)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    RangeSet,
    ConstraintList,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model
from pyomo.repn.beta.matrix import compile_block_linear_constraints

has_numpy = False
try:
    import numpy

    has_numpy = True
except:
    pass

has_scipy = False
try:
    import scipy
    import scipy.sparse

    has_scipy = True
except:
    pass


@register_model
class LP_compiled(_BaseTestModel):
    """
    A continuous linear model that is compiled into a
    MatrixConstraint object.
    """

    description = "LP_compiled"
    capabilities = set(['linear'])
    test_pickling = False
    size = (13, 22, None)

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")
        self.disable_suffix_tests = True

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.s = RangeSet(1, 12)
        model.x = Var(model.s)
        model.x[1].setlb(-1)
        model.x[1].setub(1)
        model.x[2].setlb(-1)
        model.x[2].setub(1)
        model.obj = Objective(
            expr=sum(model.x[i] * ((-1) ** (i + 1)) for i in model.x.index_set())
        )
        model.c = ConstraintList()
        # to make the variable used in the constraint match the name
        model.c.add(Constraint.Skip)
        model.c.add(Constraint.Skip)
        model.c.add(model.x[3] >= -1.0)
        model.c.add(model.x[4] <= 1.0)
        model.c.add(model.x[5] == -1.0)
        model.c.add(model.x[6] == -1.0)
        model.c.add(model.x[7] == 1.0)
        model.c.add(model.x[8] == 1.0)
        model.c.add((-1.0, model.x[9], -1.0))
        model.c.add((-1.0, model.x[10], -1.0))
        model.c.add((1.0, model.x[11], 1.0))
        model.c.add((1.0, model.x[12], 1.0))
        cdata = model.c.add((0, 1, 3))
        assert cdata.lower == 0
        assert cdata.upper == 3
        assert cdata.body() == 1
        assert not cdata.equality
        cdata = model.c.add((0, 2, 3))
        assert cdata.lower == 0
        assert cdata.upper == 3
        assert cdata.body() == 2
        assert not cdata.equality
        cdata = model.c.add((0, 1, None))
        assert cdata.lower is None
        assert cdata.upper == 1
        assert cdata.body() == 0
        assert not cdata.equality
        cdata = model.c.add((None, 0, 1))
        assert cdata.lower is None
        assert cdata.upper == 1
        assert cdata.body() == 0
        assert not cdata.equality
        cdata = model.c.add((1, 1))
        assert cdata.lower == 1
        assert cdata.upper == 1
        assert cdata.body() == 1
        assert cdata.equality

        model.fixed_var = Var()
        model.fixed_var.fix(1.0)
        cdata = model.c.add((0, 1 + model.fixed_var, 3))
        cdata = model.c.add((0, 2 + model.fixed_var, 3))
        cdata = model.c.add((0, model.fixed_var, None))
        cdata = model.c.add((None, model.fixed_var, 1))
        cdata = model.c.add((model.fixed_var, 1))

        model.c_inactive = ConstraintList()
        # to make the variable used in the constraint match the name
        model.c_inactive.add(Constraint.Skip)
        model.c_inactive.add(Constraint.Skip)
        model.c_inactive.add(model.x[3] >= -2.0)
        model.c_inactive.add(model.x[4] <= 2.0)

        compile_block_linear_constraints(model, 'Amatrix')

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        for i in model.s:
            model.x[i].value = None


if has_numpy and has_scipy:
    # TODO: we need to somehow label this as a skip rather
    #       than not defining the test class

    @register_model
    class LP_compiled_dense_kernel(LP_compiled):
        def _get_dense_data(self):
            assert has_numpy and has_scipy
            A = numpy.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ],
                dtype=float,
            )
            lb = numpy.array(
                [
                    -1.0,
                    -numpy.inf,
                    -1.0,
                    -1.0,
                    1.0,
                    1.0,
                    -1.0,
                    -1.0,
                    1.0,
                    1.0,
                    -1.0,
                    -2.0,
                    -1.0,
                    -numpy.inf,
                    0.0,
                    -2.0,
                    -3.0,
                    -1.0,
                    -numpy.inf,
                    0.0,
                    -2.0,
                    -numpy.inf,
                ]
            )
            ub = numpy.array(
                [
                    numpy.inf,
                    1.0,
                    -1.0,
                    -1.0,
                    1.0,
                    1.0,
                    -1.0,
                    -1.0,
                    1.0,
                    1.0,
                    2.0,
                    1.0,
                    numpy.inf,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    numpy.inf,
                    0.0,
                    0.0,
                    numpy.inf,
                    2.0,
                ]
            )
            eq_index = [2, 3, 4, 5, 14, 19]

            return A, lb, ub, eq_index

        def _generate_base_model(self):
            self.model = pmo.block()
            model = self.model
            model._name = self.description

            model.s = list(range(1, 13))
            model.x = pmo.variable_dict(((i, pmo.variable()) for i in model.s))
            model.x[1].lb = -1
            model.x[1].ub = 1
            model.x[2].lb = -1
            model.x[2].ub = 1
            model.obj = pmo.objective(
                expr=sum(model.x[i] * ((-1) ** (i + 1)) for i in model.s)
            )
            variable_order = [
                model.x[3],
                model.x[4],
                model.x[5],
                model.x[6],
                model.x[7],
                model.x[8],
                model.x[9],
                model.x[10],
                model.x[11],
                model.x[12],
            ]

            return variable_order

        def _generate_model(self):
            x = self._generate_base_model()
            model = self.model
            A, lb, ub, eq_index = self._get_dense_data()
            model.Amatrix = pmo.matrix_constraint(A, lb=lb, ub=ub, x=x, sparse=False)
            for i in eq_index:
                assert model.Amatrix[i].lb == model.Amatrix[i].ub
                model.Amatrix[i].rhs = model.Amatrix[i].lb

    @register_model
    class LP_compiled_sparse_kernel(LP_compiled_dense_kernel):
        def _generate_model(self):
            x = self._generate_base_model()
            model = self.model
            A, lb, ub, eq_index = self._get_dense_data()
            model.Amatrix = pmo.matrix_constraint(A, lb=lb, ub=ub, x=x, sparse=True)
            for i in eq_index:
                assert model.Amatrix[i].lb == model.Amatrix[i].ub
                model.Amatrix[i].rhs = model.Amatrix[i].lb



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pandas as pd
from os.path import join, abspath, dirname
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
    ReactorDesignExperiment,
)


def main():

    # Read in data
    file_dirname = dirname(abspath(str(__file__)))
    file_name = abspath(join(file_dirname, "reactor_data.csv"))
    data = pd.read_csv(file_name)

    # Create an experiment list
    exp_list = []
    for i in range(data.shape[0]):
        exp_list.append(ReactorDesignExperiment(data, i))

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()

    pest = parmest.Estimator(exp_list, obj_function='SSE')

    # Parameter estimation
    obj, theta = pest.theta_est()

    # Parameter estimation with bootstrap resampling
    bootstrap_theta = pest.theta_est_bootstrap(50, seed=524)

    # Plot results
    parmest.graphics.pairwise_plot(bootstrap_theta, title="Bootstrap theta")
    parmest.graphics.pairwise_plot(
        bootstrap_theta,
        theta,
        0.8,
        ["MVN", "KDE", "Rect"],
        title="Bootstrap theta with confidence regions",
    )


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for nontrivial Bounds
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import AbstractModel, Param, Var, Constraint


class Test(unittest.TestCase):
    # Test constraint bounds
    def test_constr_lower(self):
        model = AbstractModel()
        model.A = Param(default=2.0, mutable=True)
        model.B = Param(default=1.5, mutable=True)
        model.C = Param(default=2.5, mutable=True)
        model.X = Var()

        def constr_rule(model):
            return (model.A * (model.B + model.C), model.X)

        model.constr = Constraint(rule=constr_rule)

        instance = model.create_instance()
        self.assertEqual(instance.constr.lower(), 8.0)

    def test_constr_upper(self):
        model = AbstractModel()
        model.A = Param(default=2.0, mutable=True)
        model.B = Param(default=1.5, mutable=True)
        model.C = Param(default=2.5, mutable=True)
        model.X = Var()

        def constr_rule(model):
            return (model.X, model.A * (model.B + model.C))

        model.constr = Constraint(rule=constr_rule)

        instance = model.create_instance()

        self.assertEqual(instance.constr.upper(), 8.0)

    def test_constr_both(self):
        model = AbstractModel()
        model.A = Param(default=2.0, mutable=True)
        model.B = Param(default=1.5, mutable=True)
        model.C = Param(default=2.5, mutable=True)
        model.X = Var()

        def constr_rule(model):
            return (
                model.A * (model.B - model.C),
                model.X,
                model.A * (model.B + model.C),
            )

        model.constr = Constraint(rule=constr_rule)

        instance = model.create_instance()

        self.assertEqual(instance.constr.lower(), -2.0)
        self.assertEqual(instance.constr.upper(), 8.0)

    # Test variable bounds
    # JPW: Disabled until we are convinced that we want to support complex parametric expressions for variable bounds.
    def test_var_bounds(self):
        model = AbstractModel()
        model.A = Param(default=2.0, mutable=True)
        model.B = Param(default=1.5, mutable=True)
        model.C = Param(default=2.5)

        def X_bounds_rule(model):
            return (model.A * (model.B - model.C), model.A * (model.B + model.C))

        model.X = Var(bounds=X_bounds_rule)

        instance = model.create_instance()

        self.assertEqual(instance.X.lb, -2.0)
        self.assertEqual(instance.X.ub, 8.0)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
The following script can be used to run semibatch parameter estimation in
parallel and save results to files for later analysis and graphics.
Example command: mpiexec -n 4 python parallel_example.py
"""

from pyomo.common.dependencies import numpy as np, pandas as pd
from itertools import product
from os.path import join, abspath, dirname
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.semibatch.semibatch import generate_model


def main():
    # Vars to estimate
    theta_names = ['k1', 'k2', 'E1', 'E2']

    # Data, list of json file names
    data = []
    file_dirname = dirname(abspath(str(__file__)))
    for exp_num in range(10):
        file_name = abspath(join(file_dirname, 'exp' + str(exp_num + 1) + '.out'))
        data.append(file_name)

    # Note, the model already includes a 'SecondStageCost' expression
    # for sum of squared error that will be used in parameter estimation

    pest = parmest.Estimator(generate_model, data, theta_names)

    ### Parameter estimation with bootstrap resampling
    bootstrap_theta = pest.theta_est_bootstrap(100)
    bootstrap_theta.to_csv('bootstrap_theta.csv')

    ### Compute objective at theta for likelihood ratio test
    k1 = np.arange(4, 24, 3)
    k2 = np.arange(40, 160, 40)
    E1 = np.arange(29000, 32000, 500)
    E2 = np.arange(38000, 42000, 500)
    theta_vals = pd.DataFrame(list(product(k1, k2, E1, E2)), columns=theta_names)

    obj_at_theta = pest.objective_at_theta(theta_vals)
    obj_at_theta.to_csv('obj_at_theta.csv')


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base import Transformation


class AbstractTransformation(Transformation):
    """
    Base class for all model transformations that produce abstract
    models.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "abstract_transformation")
        super(AbstractTransformation, self).__init__(**kwds)


class ConcreteTransformation(Transformation):
    """
    Base class for all model transformations that produce concrete
    models.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "concrete_transformation")
        super(ConcreteTransformation, self).__init__(**kwds)


class IsomorphicTransformation(Transformation):
    """
    Base class for 'lossless' transformations for which a bijective
    mapping between optimal variable values and the optimal cost
    exists.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "isomorphic_transformation")
        super(IsomorphicTransformation, self).__init__(**kwds)


class LinearTransformation(Transformation):
    """Base class for all linear model transformations."""

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "linear_transform")
        super(LinearTransformation, self).__init__(**kwds)


class NonIsomorphicTransformation(Transformation):
    """
    Base class for 'lossy' transformations for which a bijective
    mapping between optimal variable values and the optimal cost does
    not  exist.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "isomorphic_transformation")
        super(NonIsomorphicTransformation, self).__init__(**kwds)


class NonlinearTransformation(Transformation):
    """Base class for all nonlinear model transformations."""

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "nonlinear_transform")
        super(NonlinearTransformation, self).__init__(**kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Between Steps (P-Split) reformulation for GDPs from:

J. Kronqvist, R. Misener, and C. Tsay, "Between Steps: Intermediate
Relaxations between big-M and Convex Hull Reformulations," 2021.
"""

from pyomo.core import Transformation, TransformationFactory


@TransformationFactory.register(
    'gdp.between_steps',
    doc="Reformulates a convex disjunctive model "
    "by splitting additively separable constraints"
    "on P sets of variables, then taking hull "
    "reformulation.",
)
class BetweenSteps_Transformation(Transformation):
    """
    Transform disjunctive model to equivalent MI(N)LP using the between steps
    transformation from Konqvist et al. 2021 [1].

    This transformation first calls the 'gdp.partition_disjuncts'
    transformation, resulting in an equivalent GDP with the constraints
    partitioned, and then takes the hull reformulation of that model to get
    an algebraic model.

    References
    ----------
        [1] J. Kronqvist, R. Misener, and C. Tsay, "Between Steps: Intermediate
            Relaxations between big-M and Convex Hull Reformulations," 2021.
    """

    def __init__(self):
        super(BetweenSteps_Transformation, self).__init__()

    def _apply_to(self, instance, **kwds):
        TransformationFactory('gdp.partition_disjuncts').apply_to(instance, **kwds)
        TransformationFactory('gdp.hull').apply_to(instance)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from pyomo.common import Factory
from pyomo.common.plugin_base import PluginError

logger = logging.getLogger('pyomo.core')


class UnknownDataManager:
    def __init__(self, *args, **kwds):
        #
        # The 'type' is the class type of the solver instance
        #
        self.type = kwds["type"]

    def available(self):
        return False


class DataManagerFactoryClass(Factory):
    #
    # This is the custom __call__ method
    #
    def __call__(self, _name=None, args=[], **kwds):
        if _name is None:
            return self
        _name = str(_name)
        if _name in self._cls:
            dm = self._cls[_name](**kwds)
            if not dm.available():
                raise PluginError(
                    "Cannot process data in %s files.  The following python packages need to be installed: %s"
                    % (_name, dm.requirements())
                )
        else:
            dm = UnknownDataManager(type=_name)
        return dm


DataManagerFactory = DataManagerFactoryClass('data file')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.contrib.solver.common.solution_loader import (
    SolutionLoaderBase,
    PersistentSolutionLoader,
)


class TestSolutionLoaderBase(unittest.TestCase):
    def test_member_list(self):
        expected_list = ['load_vars', 'get_primals', 'get_duals', 'get_reduced_costs']
        method_list = [
            method
            for method in dir(SolutionLoaderBase)
            if method.startswith('_') is False
        ]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_solution_loader_base(self):
        self.instance = SolutionLoaderBase()
        with self.assertRaises(NotImplementedError):
            self.instance.get_primals()
        with self.assertRaises(NotImplementedError):
            self.instance.get_duals()
        with self.assertRaises(NotImplementedError):
            self.instance.get_reduced_costs()


class TestPersistentSolutionLoader(unittest.TestCase):
    def test_member_list(self):
        expected_list = [
            'load_vars',
            'get_primals',
            'get_duals',
            'get_reduced_costs',
            'invalidate',
        ]
        method_list = [
            method
            for method in dir(PersistentSolutionLoader)
            if method.startswith('_') is False
        ]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_default_initialization(self):
        # Realistically, a solver object should be passed into this.
        # However, it works with a string. It'll just error loudly if you
        # try to run get_primals, etc.
        self.instance = PersistentSolutionLoader('ipopt')
        self.assertTrue(self.instance._valid)
        self.assertEqual(self.instance._solver, 'ipopt')

    def test_invalid(self):
        self.instance = PersistentSolutionLoader('ipopt')
        self.instance.invalidate()
        with self.assertRaises(RuntimeError):
            self.instance.get_primals()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import logging
from pyomo.common.errors import InfeasibleConstraintException, IntervalException

logger = logging.getLogger(__name__)
inf = float('inf')


class _bool_flag:
    def __init__(self, val):
        self._val = val

    def __bool__(self):
        return self._val

    def _op(self, *others):
        raise ValueError(
            f"{self._val!r} ({type(self._val).__name__}) is not a valid numeric type. "
            f"Cannot compute bounds on expression."
        )

    def __repr__(self):
        return repr(self._val)

    __float__ = _op
    __int__ = _op
    __abs__ = _op
    __neg__ = _op
    __add__ = _op
    __sub__ = _op
    __mul__ = _op
    __div__ = _op
    __pow__ = _op
    __radd__ = _op
    __rsub__ = _op
    __rmul__ = _op
    __rdiv__ = _op
    __rpow__ = _op


_true = _bool_flag(True)
_false = _bool_flag(False)


def BoolFlag(val):
    return _true if val else _false


def ineq(xl, xu, yl, yu, feasibility_tol):
    """Compute the "bounds" on an InequalityExpression

    Note this is *not* performing interval arithmetic: we are
    calculating the "bounds" on a RelationalExpression (whose domain is
    {True, False}).  Therefore we are determining if `x` can be less
    than `y`, `x` can not be less than `y`, or both.

    """
    ans = []
    if yl < xu - feasibility_tol:
        ans.append(_false)
    if xl <= yu + feasibility_tol:
        ans.append(_true)
    assert ans
    if len(ans) == 1:
        ans.append(ans[0])
    return tuple(ans)


def eq(xl, xu, yl, yu, feasibility_tol):
    """Compute the "bounds" on an EqualityExpression

    Note this is *not* performing interval arithmetic: we are
    calculating the "bounds" on a RelationalExpression (whose domain is
    {True, False}).  Therefore we are determining if `x` can be equal to
    `y`, `x` can not be equal to `y`, or both.

    """
    ans = []
    if (
        abs(xl - xu) > feasibility_tol
        or abs(yl - yu) > feasibility_tol
        or abs(xl - yl) > feasibility_tol
    ):
        ans.append(_false)
    if xl <= yu + feasibility_tol and yl <= xu + feasibility_tol:
        ans.append(_true)
    assert ans
    if len(ans) == 1:
        ans.append(ans[0])
    return tuple(ans)


def ranged(xl, xu, yl, yu, zl, zu, feasibility_tol):
    """Compute the "bounds" on a RangedExpression

    Note this is *not* performing interval arithmetic: we are
    calculating the "bounds" on a RelationalExpression (whose domain is
    {True, False}).  Therefore we are determining if `y` can be between
    `z` and `z`, `y` can be outside the range `x` and `z`, or both.

    """
    lb = ineq(xl, xu, yl, yu, feasibility_tol)
    ub = ineq(yl, yu, zl, zu, feasibility_tol)
    ans = []
    if not lb[0] or not ub[0]:
        ans.append(_false)
    if lb[1] and ub[1]:
        ans.append(_true)
    if len(ans) == 1:
        ans.append(ans[0])
    return tuple(ans)


def if_(il, iu, tl, tu, fl, fu):
    l = []
    u = []
    if iu:
        l.append(tl)
        u.append(tu)
    if not il:
        l.append(fl)
        u.append(fu)
    return min(l), max(u)


def add(xl, xu, yl, yu):
    return xl + yl, xu + yu


def sub(xl, xu, yl, yu):
    return xl - yu, xu - yl


def mul(xl, xu, yl, yu):
    lb = inf
    ub = -inf
    for i in (xl * yl, xu * yu, xu * yl, xl * yu):
        if i < lb:
            lb = i
        if i > ub:
            ub = i
        if i != i:  # math.isnan(i)
            return (-inf, inf)
    return lb, ub


def inv(xl, xu, feasibility_tol):
    """Compute the inverse of an interval

    The case where xl is very slightly positive but should be very
    slightly negative (or xu is very slightly negative but should be
    very slightly positive) should not be an issue. Suppose xu is 2 and
    xl is 1e-15 but should be -1e-15.  The bounds obtained from this
    function will be [0.5, 1e15] or [0.5, inf), depending on the value
    of feasibility_tol. The true bounds are (-inf, -1e15] U [0.5, inf),
    where U is union. The exclusion of (-inf, -1e15] should be
    acceptable. Additionally, it very important to return a non-negative
    interval when xl is non-negative.

    """
    if xu - xl <= -feasibility_tol:
        raise InfeasibleConstraintException(
            f'lower bound is greater than upper bound in inv; xl: {xl}; xu: {xu}'
        )
    elif xu <= 0 <= xl:
        # This has to return -inf to inf because it could later be multiplied by 0
        lb = -inf
        ub = inf
    elif xl < 0 < xu:
        lb = -inf
        ub = inf
    elif 0 <= xl <= feasibility_tol:
        # xu must be strictly positive
        ub = inf
        lb = 1.0 / xu
    elif xl > feasibility_tol:
        # xl and xu must be strictly positive
        ub = 1.0 / xl
        lb = 1.0 / xu
    elif -feasibility_tol <= xu <= 0:
        # xl must be strictly negative
        lb = -inf
        ub = 1.0 / xl
    elif xu < -feasibility_tol:
        # xl and xu must be strictly negative
        ub = 1.0 / xl
        lb = 1.0 / xu
    else:
        # everything else
        lb = -inf
        ub = inf
    return lb, ub


def div(xl, xu, yl, yu, feasibility_tol):
    return mul(xl, xu, *inv(yl, yu, feasibility_tol))


def power(xl, xu, yl, yu, feasibility_tol):
    """
    Compute bounds on x**y.
    """
    if xl > 0:
        # If x is always positive, things are simple. We only need to
        # worry about the sign of y.
        if yl < 0 < yu:
            lb = min(xu**yl, xl**yu)
            ub = max(xl**yl, xu**yu)
        elif yl >= 0:
            lb = min(xl**yl, xl**yu)
            ub = max(xu**yl, xu**yu)
        else:  # yu <= 0:
            lb = min(xu**yl, xu**yu)
            ub = max(xl**yl, xl**yu)
    elif xl == 0:
        if yl >= 0:
            lb = min(xl**yl, xl**yu)
            ub = max(xu**yl, xu**yu)
        elif yu <= 0:
            lb, ub = inv(
                *power(xl, xu, *sub(0, 0, yl, yu), feasibility_tol), feasibility_tol
            )
        else:
            lb1, ub1 = power(xl, xu, 0, yu, feasibility_tol)
            lb2, ub2 = power(xl, xu, yl, 0, feasibility_tol)
            lb = min(lb1, lb2)
            ub = max(ub1, ub2)
    elif yl == yu and yl == round(yl):
        # the exponent is an integer, so x can be negative
        """
        The logic here depends on several things:
        1) The sign of x
        2) The sign of y
        3) Whether y is even or odd.

        There are also special cases to avoid math domain errors.
        """
        y = yl
        if xu <= 0:
            if y < 0:
                if y % 2 == 0:
                    lb = xl**y
                    if xu == 0:
                        ub = inf
                    else:
                        ub = xu**y
                else:
                    if xu == 0:
                        lb = -inf
                        ub = inf
                    else:
                        lb = xu**y
                        ub = xl**y
            else:
                if y % 2 == 0:
                    lb = xu**y
                    ub = xl**y
                else:
                    lb = xl**y
                    ub = xu**y
        else:  # xu is positive
            if y < 0:
                if y % 2 == 0:
                    lb = min(xl**y, xu**y)
                    ub = inf
                else:
                    lb = -inf
                    ub = inf
            else:  # exponent is nonnegative
                if y % 2 == 0:
                    # xl is negative and xu is positive, so lb is 0
                    lb = 0
                    ub = max(xl**y, xu**y)
                else:
                    lb = xl**y
                    ub = xu**y
    elif yl == yu:
        # the exponent has to be fractional, so x must be positive
        if xu < 0:
            msg = 'Cannot raise a negative number to the power of {0}.\n'.format(yl)
            msg += 'The upper bound of a variable raised to the power of {0} is {1}'.format(
                yl, xu
            )
            raise InfeasibleConstraintException(msg)
        xl = 0
        lb, ub = power(xl, xu, yl, yu, feasibility_tol)
    else:
        lb = -inf
        ub = inf

    return lb, ub


def _inverse_power1(zl, zu, yl, yu, orig_xl, orig_xu, feasibility_tol):
    """z = x**y => compute bounds on x.

    First, start by computing bounds on x with

        x = exp(ln(z) / y)

    However, if y is an integer, then x can be negative, so there are
    several special cases. See the docs below.

    """
    xl, xu = log(zl, zu)
    xl, xu = div(xl, xu, yl, yu, feasibility_tol)
    xl, xu = exp(xl, xu)

    # if y is an integer, then x can be negative
    if yl == yu and yl == round(yl):  # y is a fixed integer
        y = yl
        if y == 0:
            # Anything to the power of 0 is 1, so if y is 0, then x can be anything
            # (assuming zl <= 1 <= zu, which is enforced when traversing
            # the tree in the other direction)
            xl = -inf
            xu = inf
        elif y % 2 == 0:
            """if y is even, then there are two primary cases (note that it is much
            easier to walk through these while looking at plots):

            case 1: y is positive

                x**y is convex, positive, and symmetric. The bounds on x
                depend on the lower bound of z. If zl <= 0, then xl
                should simply be -xu. However, if zl > 0, then we may be
                able to say something better. For example, if the
                original lower bound on x is positive, then we can keep
                xl computed from x = exp(ln(z) / y). Furthermore, if the
                original lower bound on x is larger than -xl computed
                from x = exp(ln(z) / y), then we can still keep the xl
                computed from x = exp(ln(z) / y). Similar logic applies
                to the upper bound of x.

            case 2: y is negative

                The ideas are similar to case 1.

            """
            if zu + feasibility_tol < 0:
                raise InfeasibleConstraintException(
                    'Infeasible. Anything to the power of {0} must be positive.'.format(
                        y
                    )
                )
            if y > 0:
                if zu <= 0:
                    _xl = 0
                    _xu = 0
                elif zl <= 0:
                    _xl = -xu
                    _xu = xu
                else:
                    if orig_xl <= -xl + feasibility_tol:
                        _xl = -xu
                    else:
                        _xl = xl
                    if orig_xu < xl - feasibility_tol:
                        _xu = -xl
                    else:
                        _xu = xu
                xl = _xl
                xu = _xu
            else:
                if zu == 0:
                    raise InfeasibleConstraintException(
                        'Infeasible. Anything to the power of {0} must be positive.'.format(
                            y
                        )
                    )
                elif zl <= 0:
                    _xl = -inf
                    _xu = inf
                else:
                    if orig_xl <= -xl + feasibility_tol:
                        _xl = -xu
                    else:
                        _xl = xl
                    if orig_xu < xl - feasibility_tol:
                        _xu = -xl
                    else:
                        _xu = xu
                xl = _xl
                xu = _xu
        else:  # y % 2 == 1
            """y is odd.

            Case 1: y is positive

                x**y is monotonically increasing. If y is positive, then
                we can can compute the bounds on x using x = z**(1/y)
                and the signs on xl and xu depend on the signs of zl and
                zu.

            Case 2: y is negative

                Again, this is easier to visualize with a plot. x**y
                approaches zero when x approaches -inf or inf.  Thus, if
                zl < 0 < zu, then no bounds can be inferred for x. If z
                is positive (zl >=0 ) then we can use the bounds
                computed from x = exp(ln(z) / y). If z is negative (zu
                <= 0), then we live in the bottom left quadrant, xl
                depends on zu, and xu depends on zl.

            """
            if y > 0:
                xl = abs(zl) ** (1.0 / y)
                xl = math.copysign(xl, zl)
                xu = abs(zu) ** (1.0 / y)
                xu = math.copysign(xu, zu)
            else:
                if zl >= 0:
                    pass
                elif zu <= 0:
                    if zu == 0:
                        xl = -inf
                    else:
                        xl = -abs(zu) ** (1.0 / y)
                    if zl == 0:
                        xu = -inf
                    else:
                        xu = -abs(zl) ** (1.0 / y)
                else:
                    xl = -inf
                    xu = inf

    return xl, xu


def _inverse_power2(zl, zu, xl, xu, feasiblity_tol):
    """z = x**y => compute bounds on y
    y = ln(z) / ln(x)

    This function assumes the exponent can be fractional, so x must be
    positive. This method should not be called if the exponent is an
    integer.

    """
    if xu <= 0:
        raise IntervalException(
            'Cannot raise a negative variable to a fractional power.'
        )
    if (xl > 0 and zu <= 0) or (xl >= 0 and zu < 0):
        raise InfeasibleConstraintException(
            'A positive variable raised to the power of anything must be positive.'
        )
    lba, uba = log(zl, zu)
    lbb, ubb = log(xl, xu)
    yl, yu = div(lba, uba, lbb, ubb, feasiblity_tol)
    return yl, yu


def interval_abs(xl, xu):
    abs_xl = abs(xl)
    abs_xu = abs(xu)
    if xl <= 0 and 0 <= xu:
        res_lb = 0
        res_ub = max(abs_xl, abs_xu)
    else:
        res_lb = min(abs_xl, abs_xu)
        res_ub = max(abs_xl, abs_xu)
    return res_lb, res_ub


def _inverse_abs(zl, zu):
    if zl < 0:
        zl = 0
    if zu < 0:
        zu = 0
    xu = max(zl, zu)
    xl = -xu
    return xl, xu


def exp(xl, xu):
    try:
        lb = math.exp(xl)
    except OverflowError:
        lb = math.inf
    try:
        ub = math.exp(xu)
    except OverflowError:
        ub = math.inf
    return lb, ub


def log(xl, xu):
    if xl > 0:
        lb = math.log(xl)
    else:
        lb = -inf
    if xu > 0:
        ub = math.log(xu)
    else:
        ub = -inf
    return lb, ub


def log10(xl, xu):
    if xl > 0:
        lb = math.log10(xl)
    else:
        lb = -inf
    if xu > 0:
        ub = math.log10(xu)
    else:
        ub = -inf
    return lb, ub


def sin(xl, xu):
    """

    Parameters
    ----------
    xl: float
    xu: float

    Returns
    -------
    lb: float
    ub: float
    """

    # if there is a minimum between xl and xu, then the lower bound is
    # -1. Minimums occur at 2*pi*n - pi/2 find the minimum value of i
    # such that 2*pi*i - pi/2 >= xl. Then round i up. If 2*pi*i - pi/2
    # is still less than or equal to xu, then there is a minimum between
    # xl and xu. Thus the lb is -1. Otherwise, the minimum occurs at
    # either xl or xu
    if xl <= -inf or xu >= inf:
        return -1, 1
    pi = math.pi
    i = (xl + pi / 2) / (2 * pi)
    i = math.ceil(i)
    x_at_min = 2 * pi * i - pi / 2
    if x_at_min <= xu:
        lb = -1
    else:
        lb = min(math.sin(xl), math.sin(xu))

    # if there is a maximum between xl and xu, then the upper bound is
    # 1. Maximums occur at 2*pi*n + pi/2
    i = (xu - pi / 2) / (2 * pi)
    i = math.floor(i)
    x_at_max = 2 * pi * i + pi / 2
    if x_at_max >= xl:
        ub = 1
    else:
        ub = max(math.sin(xl), math.sin(xu))

    return lb, ub


def cos(xl, xu):
    """

    Parameters
    ----------
    xl: float
    xu: float

    Returns
    -------
    lb: float
    ub: float
    """

    # if there is a minimum between xl and xu, then the lower bound is
    # -1. Minimums occur at 2*pi*n - pi find the minimum value of i such
    # that 2*pi*i - pi >= xl. Then round i up. If 2*pi*i - pi/2 is still
    # less than or equal to xu, then there is a minimum between xl and
    # xu. Thus the lb is -1. Otherwise, the minimum occurs at either xl
    # or xu
    if xl <= -inf or xu >= inf:
        return -1, 1
    pi = math.pi
    i = (xl + pi) / (2 * pi)
    i = math.ceil(i)
    x_at_min = 2 * pi * i - pi
    if x_at_min <= xu:
        lb = -1
    else:
        lb = min(math.cos(xl), math.cos(xu))

    # if there is a maximum between xl and xu, then the upper bound is
    # 1. Maximums occur at 2*pi*n
    i = (xu) / (2 * pi)
    i = math.floor(i)
    x_at_max = 2 * pi * i
    if x_at_max >= xl:
        ub = 1
    else:
        ub = max(math.cos(xl), math.cos(xu))

    return lb, ub


def tan(xl, xu):
    """

    Parameters
    ----------
    xl: float
    xu: float

    Returns
    -------
    lb: float
    ub: float
    """

    # tan goes to -inf and inf at every pi*i + pi/2 (integer i). If one
    # of these values is between xl and xu, then the lb is -inf and the
    # ub is inf. Otherwise the minimum occurs at xl and the maximum
    # occurs at xu.  find the minimum value of i such that pi*i + pi/2
    # >= xl. Then round i up. If pi*i + pi/2 is still less than or equal
    # to xu, then there is an undefined point between xl and xu.
    if xl <= -inf or xu >= inf:
        return -inf, inf
    pi = math.pi
    i = (xl - pi / 2) / (pi)
    i = math.ceil(i)
    x_at_undefined = pi * i + pi / 2
    if x_at_undefined <= xu:
        lb = -inf
        ub = inf
    else:
        lb = math.tan(xl)
        ub = math.tan(xu)

    return lb, ub


def asin(xl, xu, yl, yu, feasibility_tol):
    """
    y = asin(x); propagate bounds from x to y
    x = sin(y)

    Parameters
    ----------
    xl: float
    xu: float
    yl: float
    yu: float

    Returns
    -------
    lb: float
    ub: float
    """
    if xl < -1:
        xl = -1
    if xu > 1:
        xu = 1

    pi = math.pi

    if yl <= -inf:
        lb = yl
    elif xl <= math.sin(yl) <= xu:
        # if sin(yl) >= xl then yl satisfies the bounds on x, and the
        # lower bound of y cannot be improved
        lb = yl
    elif math.sin(yl) < xl:
        """we can only push yl up from its current value to the next lowest
        value such that xl = sin(y). In other words, we need to

        min y
        s.t.
            xl = sin(y)
            y >= yl

        globally.

        """
        # first find the next minimum of x = sin(y). Minimums occur at y
        # = 2*pi*n - pi/2 for integer n.
        i = (yl + pi / 2) / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1 - pi / 2
        i2 = 2 * pi * i2 - pi / 2
        # now find the next value of y such that xl = sin(y). This can
        # be computed by a distance from the minimum (i).
        y_tmp = math.asin(xl)  # this will give me a value between -pi/2 and pi/2
        dist = y_tmp - (-pi / 2)
        # this is the distance between the minimum of the sin function
        # and a value that satisfies xl = sin(y)
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2
    else:
        # sin(yl) > xu
        i = (yl - pi / 2) / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1 + pi / 2
        i2 = 2 * pi * i2 + pi / 2
        y_tmp = math.asin(xu)
        dist = pi / 2 - y_tmp
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2

    # use the same logic for the maximum
    if yu >= inf:
        ub = yu
    elif xl <= math.sin(yu) <= xu:
        ub = yu
    elif math.sin(yu) > xu:
        i = (yu - pi / 2) / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1 + pi / 2
        i2 = 2 * pi * i2 + pi / 2
        y_tmp = math.asin(xu)
        dist = pi / 2 - y_tmp
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2
    else:
        # math.sin(yu) < xl
        i = (yu + pi / 2) / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1 - pi / 2
        i2 = 2 * pi * i2 - pi / 2
        y_tmp = math.asin(xl)
        dist = y_tmp - (-pi / 2)
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2

    return lb, ub


def acos(xl, xu, yl, yu, feasibility_tol):
    """
    y = acos(x); propagate bounds from x to y
    x = cos(y)

    Parameters
    ----------
    xl: float
    xu: float
    yl: float
    yu: float

    Returns
    -------
    lb: float
    ub: float
    """
    if xl < -1:
        xl = -1
    if xu > 1:
        xu = 1

    pi = math.pi

    if yl <= -inf:
        lb = yl
    elif xl <= math.cos(yl) <= xu:
        # if xl <= cos(yl) <= xu then yl satisfies the bounds on x, and
        # the lower bound of y cannot be improved
        lb = yl
    elif math.cos(yl) < xl:
        """we can only push yl up from its current value to the next lowest
        value such that xl = cos(y). In other words, we need to

        min y
        s.t.
            xl = cos(y)
            y >= yl

        globally.

        """
        # first find the next minimum of x = cos(y). Minimums occur at y
        # = 2*pi*n - pi for integer n.
        i = (yl + pi) / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1 - pi
        i2 = 2 * pi * i2 - pi
        # now find the next value of y such that xl = cos(y). This can
        # be computed by a distance from the minimum (i).
        y_tmp = math.acos(xl)  # this will give me a value between 0 and pi
        dist = pi - y_tmp
        # this is the distance between the minimum of the sin function
        # and a value that satisfies xl = sin(y)
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2
    else:
        # cos(yl) > xu
        # first find the next maximum of x = cos(y).
        i = yl / (2 * pi)
        i1 = math.floor(i)
        i2 = math.ceil(i)
        i1 = 2 * pi * i1
        i2 = 2 * pi * i2
        y_tmp = math.acos(xu)
        dist = y_tmp
        lb1 = i1 + dist
        lb2 = i2 + dist
        if lb1 >= yl - feasibility_tol:
            lb = lb1
        else:
            lb = lb2

    # use the same logic for the maximum
    if yu >= inf:
        ub = yu
    elif xl <= math.cos(yu) <= xu:
        ub = yu
    elif math.cos(yu) > xu:
        i = yu / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1
        i2 = 2 * pi * i2
        y_tmp = math.acos(xu)
        dist = y_tmp
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2
    else:
        # math.cos(yu) < xl
        i = (yu + pi) / (2 * pi)
        i1 = math.ceil(i)
        i2 = math.floor(i)
        i1 = 2 * pi * i1 - pi
        i2 = 2 * pi * i2 - pi
        y_tmp = math.acos(xl)
        dist = pi - y_tmp
        ub1 = i1 - dist
        ub2 = i2 - dist
        if ub1 <= yu + feasibility_tol:
            ub = ub1
        else:
            ub = ub2

    return lb, ub


def atan(xl, xu, yl, yu):
    """
    y = atan(x); propagate bounds from x to y
    x = tan(y)

    Parameters
    ----------
    xl: float
    xu: float
    yl: float
    yu: float

    Returns
    -------
    lb: float
    ub: float
    """

    pi = math.pi

    # tan goes to -inf and inf at every pi*i + pi/2 (integer i).
    if xl <= -inf or yl <= -inf:
        lb = yl
    else:
        i = (yl - pi / 2) / pi
        i = math.floor(i)
        i = pi * i + pi / 2
        y_tmp = math.atan(xl)
        dist = y_tmp - (-pi / 2)
        lb = i + dist

    if xu >= inf or yu >= inf:
        ub = yu
    else:
        i = (yu - pi / 2) / pi
        i = math.ceil(i)
        i = pi * i + pi / 2
        y_tmp = math.atan(xu)
        dist = pi / 2 - y_tmp
        ub = i - dist

    if yl > lb:
        lb = yl
    if yu < ub:
        ub = yu

    return lb, ub
